@@ -1,23 +1,26 @@
+import asyncio
 import inspect
 import json
+import logging
 import warnings
 import zlib
-import asyncio
-
-from collections.abc import Iterable
 from datetime import datetime, timedelta, timezone
 from enum import Enum
-from redis import WatchError
-from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union
 from uuid import uuid4
 
+from redis import WatchError
+
+from .defaults import CALLBACK_TIMEOUT, UNSERIALIZABLE_RETURN_VALUE_PAYLOAD
+from .timeouts import BaseDeathPenalty, JobTimeoutException
 
 if TYPE_CHECKING:
-    from .results import Result
-    from .queue import Queue
     from redis import Redis
     from redis.client import Pipeline
 
+    from .queue import Queue
+    from .results import Result
+
 from .connections import resolve_connection
 from .exceptions import DeserializationError, InvalidJobOperation, NoSuchJobError
 from .local import LocalStack
@@ -36,6 +39,8 @@ from .utils import (
     utcnow,
 )
 
+logger = logging.getLogger("rq.job")
+
 
 class JobStatus(str, Enum):
     """The Status of Job within its lifecycle at any given time."""
@@ -148,13 +153,14 @@ class Job:
         depends_on: Optional[JobDependencyType] = None,
         timeout: Optional[int] = None,
         id: Optional[str] = None,
-        origin=None,
+        origin: str = '',
         meta: Optional[Dict[str, Any]] = None,
         failure_ttl: Optional[int] = None,
         serializer=None,
         *,
-        on_success: Optional[Callable[..., Any]] = None,
-        on_failure: Optional[Callable[..., Any]] = None
+        on_success: Optional[Union['Callback', Callable[..., Any]]] = None,  # Callable is deprecated
+        on_failure: Optional[Union['Callback', Callable[..., Any]]] = None,  # Callable is deprecated
+        on_stopped: Optional[Union['Callback', Callable[..., Any]]] = None,  # Callable is deprecated
     ) -> 'Job':
         """Creates a new Job instance for the given function, arguments, and
         keyword arguments.
@@ -163,8 +169,8 @@ class Job:
             func (FunctionReference): The function/method/callable for the Job. This can be
                 a reference to a concrete callable or a string representing the  path of function/method to be
                 imported. Effectively this is the only required attribute when creating a new Job.
-            args (Union[List[Any], Optional[Tuple]], optional): A Tuple / List of positional arguments to pass the callable.
-                Defaults to None, meaning no args being passed.
+            args (Union[List[Any], Optional[Tuple]], optional): A Tuple / List of positional arguments to pass the
+                callable.  Defaults to None, meaning no args being passed.
             kwargs (Optional[Dict], optional): A Dictionary of keyword arguments to pass the callable.
                 Defaults to None, meaning no kwargs being passed.
             connection (Optional[Redis], optional): The Redis connection to use. Defaults to None.
@@ -175,26 +181,32 @@ class Job:
             status (JobStatus, optional): The Job Status. Defaults to None.
             description (Optional[str], optional): The Job Description. Defaults to None.
             depends_on (Union['Dependency', List[Union['Dependency', 'Job']]], optional): What the jobs depends on.
-                This accepts a variaty of different arguments including a `Dependency`, a list of `Dependency` or a `Job`
-                list of `Job`. Defaults to None.
-            timeout (Optional[int], optional): The amount of time in seconds that should be a hardlimit for a job execution. Defaults to None.
+                This accepts a variaty of different arguments including a `Dependency`, a list of `Dependency` or a
+                `Job` list of `Job`. Defaults to None.
+            timeout (Optional[int], optional): The amount of time in seconds that should be a hardlimit for a job
+                execution. Defaults to None.
             id (Optional[str], optional): An Optional ID (str) for the Job. Defaults to None.
             origin (Optional[str], optional): The queue of origin. Defaults to None.
-            meta (Optional[Dict[str, Any]], optional): Custom metadata about the job, takes a dictioanry. Defaults to None.
-            failure_ttl (Optional[int], optional): THe time to live in seconds for failed-jobs information. Defaults to None.
+            meta (Optional[Dict[str, Any]], optional): Custom metadata about the job, takes a dictioanry.
+                Defaults to None.
+            failure_ttl (Optional[int], optional): THe time to live in seconds for failed-jobs information.
+                Defaults to None.
             serializer (Optional[str], optional): The serializer class path to use. Should be a string with the import
                 path for the serializer to use. eg. `mymodule.myfile.MySerializer` Defaults to None.
-            on_success (Optional[Callable[..., Any]], optional): A callback function, should be a callable to run
-                when/if the Job finishes sucessfully. Defaults to None.
-            on_failure (Optional[Callable[..., Any]], optional): A callback function, should be a callable to run
-                when/if the Job fails. Defaults to None.
+            on_success (Optional[Union['Callback', Callable[..., Any]]], optional): A callback to run when/if the Job
+                finishes sucessfully. Defaults to None. Passing a callable is deprecated.
+            on_failure (Optional[Union['Callback', Callable[..., Any]]], optional): A callback to run when/if the Job
+                fails. Defaults to None. Passing a callable is deprecated.
+            on_stopped (Optional[Union['Callback', Callable[..., Any]]], optional): A callback to run when/if the Job
+                is stopped. Defaults to None. Passing a callable is deprecated.
 
         Raises:
             TypeError: If `args` is not a tuple/list
             TypeError: If `kwargs` is not a dict
             TypeError: If the `func` is something other than a string or a Callable reference
-            ValueError: If `on_failure` is not a function
-            ValueError: If `on_success` is not a function
+            ValueError: If `on_failure` is not a Callback or function or string
+            ValueError: If `on_success` is not a Callback or function or string
+            ValueError: If `on_stopped` is not a Callback or function or string
 
         Returns:
             Job: A job instance.
@@ -213,7 +225,7 @@ class Job:
         if id is not None:
             job.set_id(id)
 
-        if origin is not None:
+        if origin:
             job.origin = origin
 
         # Set the core job tuple properties
@@ -234,14 +246,34 @@ class Job:
         job._kwargs = kwargs
 
         if on_success:
-            if not inspect.isfunction(on_success) and not inspect.isbuiltin(on_success):
-                raise ValueError('on_success callback must be a function')
-            job._success_callback_name = '{0}.{1}'.format(on_success.__module__, on_success.__qualname__)
+            if not isinstance(on_success, Callback):
+                warnings.warn(
+                    'Passing a string or function for `on_success` is deprecated, pass `Callback` instead',
+                    DeprecationWarning,
+                )
+                on_success = Callback(on_success)  # backward compatibility
+            job._success_callback_name = on_success.name
+            job._success_callback_timeout = on_success.timeout
 
         if on_failure:
-            if not inspect.isfunction(on_failure) and not inspect.isbuiltin(on_failure):
-                raise ValueError('on_failure callback must be a function')
-            job._failure_callback_name = '{0}.{1}'.format(on_failure.__module__, on_failure.__qualname__)
+            if not isinstance(on_failure, Callback):
+                warnings.warn(
+                    'Passing a string or function for `on_failure` is deprecated, pass `Callback` instead',
+                    DeprecationWarning,
+                )
+                on_failure = Callback(on_failure)  # backward compatibility
+            job._failure_callback_name = on_failure.name
+            job._failure_callback_timeout = on_failure.timeout
+
+        if on_stopped:
+            if not isinstance(on_stopped, Callback):
+                warnings.warn(
+                    'Passing a string or function for `on_stopped` is deprecated, pass `Callback` instead',
+                    DeprecationWarning,
+                )
+                on_stopped = Callback(on_stopped)  # backward compatibility
+            job._stopped_callback_name = on_stopped.name
+            job._stopped_callback_timeout = on_stopped.timeout
 
         # Extra meta data
         job.description = description or job.get_call_string()
@@ -254,12 +286,17 @@ class Job:
 
         # dependency could be job instance or id, or iterable thereof
         if depends_on is not None:
-            if isinstance(depends_on, Dependency):
-                job.enqueue_at_front = depends_on.enqueue_at_front
-                job.allow_dependency_failures = depends_on.allow_failure
-                depends_on_list = depends_on.dependencies
-            else:
-                depends_on_list = ensure_list(depends_on)
+            depends_on = ensure_list(depends_on)
+            depends_on_list = []
+            for depends_on_item in depends_on:
+                if isinstance(depends_on_item, Dependency):
+                    # If a Dependency has enqueue_at_front or allow_failure set to True, these behaviors are used for
+                    # all dependencies.
+                    job.enqueue_at_front = job.enqueue_at_front or depends_on_item.enqueue_at_front
+                    job.allow_dependency_failures = job.allow_dependency_failures or depends_on_item.allow_failure
+                    depends_on_list.extend(depends_on_item.dependencies)
+                else:
+                    depends_on_list.extend(ensure_list(depends_on_item))
             job._dependency_ids = [dep.id if isinstance(dep, Job) else dep for dep in depends_on_list]
 
         return job
@@ -287,7 +324,8 @@ class Job:
             status (JobStatus): The Job Status
         """
         if refresh:
-            self._status = as_text(self.connection.hget(self.key, 'status'))
+            status = self.connection.hget(self.key, 'status')
+            self._status = as_text(status) if status else None
         return self._status
 
     def set_status(self, status: JobStatus, pipeline: Optional['Pipeline'] = None) -> None:
@@ -396,6 +434,13 @@ class Job:
 
         return self._success_callback
 
+    @property
+    def success_callback_timeout(self) -> int:
+        if self._success_callback_timeout is None:
+            return CALLBACK_TIMEOUT
+
+        return self._success_callback_timeout
+
     @property
     def failure_callback(self):
         if self._failure_callback is UNEVALUATED:
@@ -406,6 +451,30 @@ class Job:
 
         return self._failure_callback
 
+    @property
+    def failure_callback_timeout(self) -> int:
+        if self._failure_callback_timeout is None:
+            return CALLBACK_TIMEOUT
+
+        return self._failure_callback_timeout
+
+    @property
+    def stopped_callback(self):
+        if self._stopped_callback is UNEVALUATED:
+            if self._stopped_callback_name:
+                self._stopped_callback = import_attribute(self._stopped_callback_name)
+            else:
+                self._stopped_callback = None
+
+        return self._stopped_callback
+
+    @property
+    def stopped_callback_timeout(self) -> int:
+        if self._stopped_callback_timeout is None:
+            return CALLBACK_TIMEOUT
+
+        return self._stopped_callback_timeout
+
     def _deserialize_data(self):
         """Deserializes the Job `data` into a tuple.
         This includes the `_func_name`, `_instance`, `_args` and `_kwargs`
@@ -468,7 +537,7 @@ class Job:
         self._data = UNEVALUATED
 
     @property
-    def args(self):
+    def args(self) -> tuple:
         if self._args is UNEVALUATED:
             self._deserialize_data()
         return self._args
@@ -500,9 +569,10 @@ class Job:
         Returns:
             job_exists (bool): Whether the Job exists
         """
-        conn = resolve_connection(connection)
+        if not connection:
+            connection = resolve_connection()
         job_key = cls.key_for(job_id)
-        job_exists = conn.exists(job_key)
+        job_exists = connection.exists(job_key)
         return bool(job_exists)
 
     @classmethod
@@ -555,7 +625,10 @@ class Job:
         return jobs
 
     def __init__(self, id: Optional[str] = None, connection: Optional['Redis'] = None, serializer=None):
-        self.connection = resolve_connection(connection)
+        if connection:
+            self.connection = connection
+        else:
+            self.connection = resolve_connection()
         self._id = id
         self.created_at = utcnow()
         self._data = UNEVALUATED
@@ -567,21 +640,26 @@ class Job:
         self._success_callback = UNEVALUATED
         self._failure_callback_name = None
         self._failure_callback = UNEVALUATED
+        self._stopped_callback_name = None
+        self._stopped_callback = UNEVALUATED
         self.description: Optional[str] = None
-        self.origin: Optional[str] = None
+        self.origin: str = ''
         self.enqueued_at: Optional[datetime] = None
         self.started_at: Optional[datetime] = None
         self.ended_at: Optional[datetime] = None
         self._result = None
         self._exc_info = None
         self.timeout: Optional[float] = None
+        self._success_callback_timeout: Optional[int] = None
+        self._failure_callback_timeout: Optional[int] = None
+        self._stopped_callback_timeout: Optional[int] = None
         self.result_ttl: Optional[int] = None
         self.failure_ttl: Optional[int] = None
         self.ttl: Optional[int] = None
         self.worker_name: Optional[str] = None
         self._status = None
         self._dependency_ids: List[str] = []
-        self.meta: Optional[Dict] = {}
+        self.meta: Dict = {}
         self.serializer = resolve_serializer(serializer)
         self.retries_left: Optional[int] = None
         self.retry_intervals: Optional[List[int]] = None
@@ -841,9 +919,9 @@ class Job:
             self.data = raw_data
 
         self.created_at = str_to_date(obj.get('created_at'))
-        self.origin = as_text(obj.get('origin'))
+        self.origin = as_text(obj.get('origin')) if obj.get('origin') else ''
         self.worker_name = obj.get('worker_name').decode() if obj.get('worker_name') else None
-        self.description = as_text(obj.get('description'))
+        self.description = as_text(obj.get('description')) if obj.get('description') else None
         self.enqueued_at = str_to_date(obj.get('enqueued_at'))
         self.started_at = str_to_date(obj.get('started_at'))
         self.ended_at = str_to_date(obj.get('ended_at'))
@@ -853,7 +931,7 @@ class Job:
             try:
                 self._result = self.serializer.loads(result)
             except Exception:
-                self._result = "Unserializable return value"
+                self._result = UNSERIALIZABLE_RETURN_VALUE_PAYLOAD
         self.timeout = parse_timeout(obj.get('timeout')) if obj.get('timeout') else None
         self.result_ttl = int(obj.get('result_ttl')) if obj.get('result_ttl') else None
         self.failure_ttl = int(obj.get('failure_ttl')) if obj.get('failure_ttl') else None
@@ -862,9 +940,21 @@ class Job:
         if obj.get('success_callback_name'):
             self._success_callback_name = obj.get('success_callback_name').decode()
 
+        if 'success_callback_timeout' in obj:
+            self._success_callback_timeout = int(obj.get('success_callback_timeout'))
+
         if obj.get('failure_callback_name'):
             self._failure_callback_name = obj.get('failure_callback_name').decode()
 
+        if 'failure_callback_timeout' in obj:
+            self._failure_callback_timeout = int(obj.get('failure_callback_timeout'))
+
+        if obj.get('stopped_callback_name'):
+            self._stopped_callback_name = obj.get('stopped_callback_name').decode()
+
+        if 'stopped_callback_timeout' in obj:
+            self._stopped_callback_timeout = int(obj.get('stopped_callback_timeout'))
+
         dep_ids = obj.get('dependency_ids')
         dep_id = obj.get('dependency_id')  # for backwards compatibility
         self._dependency_ids = json.loads(dep_ids.decode()) if dep_ids else [dep_id.decode()] if dep_id else []
@@ -872,7 +962,10 @@ class Job:
         self.allow_dependency_failures = bool(int(allow_failures)) if allow_failures else None
         self.enqueue_at_front = bool(int(obj['enqueue_at_front'])) if 'enqueue_at_front' in obj else None
         self.ttl = int(obj.get('ttl')) if obj.get('ttl') else None
-        self.meta = self.serializer.loads(obj.get('meta')) if obj.get('meta') else {}
+        try:
+            self.meta = self.serializer.loads(obj.get('meta')) if obj.get('meta') else {}
+        except Exception:  # depends on the serializer
+            self.meta = {'unserialized': obj.get('meta', {})}
 
         self.retries_left = int(obj.get('retries_left')) if obj.get('retries_left') else None
         if obj.get('retry_intervals'):
@@ -916,6 +1009,7 @@ class Job:
             'data': zlib.compress(self.data),
             'success_callback_name': self._success_callback_name if self._success_callback_name else '',
             'failure_callback_name': self._failure_callback_name if self._failure_callback_name else '',
+            'stopped_callback_name': self._stopped_callback_name if self._stopped_callback_name else '',
             'started_at': utcformat(self.started_at) if self.started_at else '',
             'ended_at': utcformat(self.ended_at) if self.ended_at else '',
             'last_heartbeat': utcformat(self.last_heartbeat) if self.last_heartbeat else '',
@@ -926,7 +1020,7 @@ class Job:
             obj['retries_left'] = self.retries_left
         if self.retry_intervals is not None:
             obj['retry_intervals'] = json.dumps(self.retry_intervals)
-        if self.origin is not None:
+        if self.origin:
             obj['origin'] = self.origin
         if self.description is not None:
             obj['description'] = self.description
@@ -942,6 +1036,12 @@ class Job:
             obj['exc_info'] = zlib.compress(str(self._exc_info).encode('utf-8'))
         if self.timeout is not None:
             obj['timeout'] = self.timeout
+        if self._success_callback_timeout is not None:
+            obj['success_callback_timeout'] = self._success_callback_timeout
+        if self._failure_callback_timeout is not None:
+            obj['failure_callback_timeout'] = self._failure_callback_timeout
+        if self._stopped_callback_timeout is not None:
+            obj['stopped_callback_timeout'] = self._stopped_callback_timeout
         if self.result_ttl is not None:
             obj['result_ttl'] = self.result_ttl
         if self.failure_ttl is not None:
@@ -1030,8 +1130,8 @@ class Job:
         """
         if self.is_canceled:
             raise InvalidJobOperation("Cannot cancel already canceled job: {}".format(self.get_id()))
-        from .registry import CanceledJobRegistry
         from .queue import Queue
+        from .registry import CanceledJobRegistry
 
         pipe = pipeline or self.connection.pipeline()
 
@@ -1303,6 +1403,44 @@ class Job:
             self.origin, connection=self.connection, job_class=self.__class__, serializer=self.serializer
         )
 
+    def execute_success_callback(self, death_penalty_class: Type[BaseDeathPenalty], result: Any):
+        """Executes success_callback for a job.
+        with timeout .
+
+        Args:
+            death_penalty_class (Type[BaseDeathPenalty]): The penalty class to use for timeout
+            result (Any): The job's result.
+        """
+        if not self.success_callback:
+            return
+
+        logger.debug('Running success callbacks for %s', self.id)
+        with death_penalty_class(self.success_callback_timeout, JobTimeoutException, job_id=self.id):
+            self.success_callback(self, self.connection, result)
+
+    def execute_failure_callback(self, death_penalty_class: Type[BaseDeathPenalty], *exc_info):
+        """Executes failure_callback with possible timeout"""
+        if not self.failure_callback:
+            return
+
+        logger.debug('Running failure callbacks for %s', self.id)
+        try:
+            with death_penalty_class(self.failure_callback_timeout, JobTimeoutException, job_id=self.id):
+                self.failure_callback(self, self.connection, *exc_info)
+        except Exception:  # noqa
+            logger.exception(f'Job {self.id}: error while executing failure callback')
+            raise
+
+    def execute_stopped_callback(self, death_penalty_class: Type[BaseDeathPenalty]):
+        """Executes stopped_callback with possible timeout"""
+        logger.debug('Running stopped callbacks for %s', self.id)
+        try:
+            with death_penalty_class(self.stopped_callback_timeout, JobTimeoutException, job_id=self.id):
+                self.stopped_callback(self, self.connection)
+        except Exception:  # noqa
+            logger.exception(f'Job {self.id}: error while executing stopped callback')
+            raise
+
     def _handle_success(self, result_ttl: int, pipeline: 'Pipeline'):
         """Saves and cleanup job after successful execution"""
         # self.log.debug('Setting job %s status to finished', job.id)
@@ -1317,9 +1455,8 @@ class Job:
         # for backward compatibility
         if self.supports_redis_streams:
             from .results import Result
-            Result.create(
-                self, Result.Type.SUCCESSFUL, return_value=self._result, ttl=result_ttl, pipeline=pipeline
-            )
+
+            Result.create(self, Result.Type.SUCCESSFUL, return_value=self._result, ttl=result_ttl, pipeline=pipeline)
 
         if result_ttl != 0:
             finished_job_registry = self.finished_job_registry
@@ -1339,6 +1476,7 @@ class Job:
         )
         if self.supports_redis_streams:
             from .results import Result
+
             Result.create_failure(self, self.failure_ttl, exc_string=exc_string, pipeline=pipeline)
 
     def get_retry_interval(self) -> int:
@@ -1372,7 +1510,7 @@ class Job:
             self.set_status(JobStatus.SCHEDULED)
             queue.schedule_job(self, scheduled_datetime, pipeline=pipeline)
         else:
-            queue.enqueue_job(self, pipeline=pipeline)
+            queue._enqueue_job(self, pipeline=pipeline)
 
     def register_dependency(self, pipeline: Optional['Pipeline'] = None):
         """Jobs may have dependencies. Jobs are enqueued only if the jobs they
@@ -1502,3 +1640,18 @@ class Retry:
 
         self.max = max
         self.intervals = intervals
+
+
+class Callback:
+    def __init__(self, func: Union[str, Callable[..., Any]], timeout: Optional[Any] = None):
+        if not isinstance(func, str) and not inspect.isfunction(func) and not inspect.isbuiltin(func):
+            raise ValueError('Callback `func` must be a string or function')
+
+        self.func = func
+        self.timeout = parse_timeout(timeout) if timeout else CALLBACK_TIMEOUT
+
+    @property
+    def name(self) -> str:
+        if isinstance(self.func, str):
+            return self.func
+        return '{0}.{1}'.format(self.func.__module__, self.func.__qualname__)
