@@ -4,29 +4,29 @@ import traceback
 import uuid
 import warnings
 from collections import namedtuple
-from datetime import datetime, timezone, timedelta
+from datetime import datetime, timedelta, timezone
 from functools import total_ordering
-from typing import TYPE_CHECKING, Dict, List, Any, Callable, Optional, Tuple, Type, Union
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, Union
 
 from redis import WatchError
 
+from .timeouts import BaseDeathPenalty, UnixSignalDeathPenalty
+
 if TYPE_CHECKING:
     from redis import Redis
     from redis.client import Pipeline
+
     from .job import Retry
 
-from .utils import as_text
 from .connections import resolve_connection
 from .defaults import DEFAULT_RESULT_TTL
+from .dependency import Dependency
 from .exceptions import DequeueTimeout, NoSuchJobError
-from .job import Job, JobStatus
-from .types import FunctionReferenceType, JobDependencyType
+from .job import Callback, Job, JobStatus
+from .logutils import blue, green
 from .serializers import resolve_serializer
-from .utils import backend_class, get_version, import_attribute, make_colorizer, parse_timeout, utcnow, compact
-
-green = make_colorizer('darkgreen')
-yellow = make_colorizer('darkyellow')
-blue = make_colorizer('darkblue')
+from .types import FunctionReferenceType, JobDependencyType
+from .utils import as_text, backend_class, compact, get_version, import_attribute, parse_timeout, utcnow
 
 logger = logging.getLogger("rq.queue")
 
@@ -43,12 +43,14 @@ class EnqueueData(
             "ttl",
             "failure_ttl",
             "description",
+            "depends_on",
             "job_id",
             "at_front",
             "meta",
             "retry",
             "on_success",
             "on_failure",
+            "on_stopped",
         ],
     )
 ):
@@ -62,13 +64,18 @@ class EnqueueData(
 @total_ordering
 class Queue:
     job_class: Type['Job'] = Job
+    death_penalty_class: Type[BaseDeathPenalty] = UnixSignalDeathPenalty
     DEFAULT_TIMEOUT: int = 180  # Default timeout seconds.
     redis_queue_namespace_prefix: str = 'rq:queue:'
     redis_queues_keys: str = 'rq:queues'
 
     @classmethod
     def all(
-            cls, connection: Optional['Redis'] = None, job_class: Optional[Type['Job']] = None, serializer=None
+        cls,
+        connection: Optional['Redis'] = None,
+        job_class: Optional[Type['Job']] = None,
+        serializer=None,
+        death_penalty_class: Optional[Type[BaseDeathPenalty]] = None,
     ) -> List['Queue']:
         """Returns an iterable of all Queues.
 
@@ -76,15 +83,20 @@ class Queue:
             connection (Optional[Redis], optional): The Redis Connection. Defaults to None.
             job_class (Optional[Job], optional): The Job class to use. Defaults to None.
             serializer (optional): The serializer to use. Defaults to None.
+            death_penalty_class (Optional[Job], optional): The Death Penalty class to use. Defaults to None.
 
         Returns:
             queues (List[Queue]): A list of all queues.
         """
-        connection = resolve_connection(connection)
+        connection = connection or resolve_connection()
 
-        def to_queue(queue_key):
+        def to_queue(queue_key: Union[bytes, str]):
             return cls.from_queue_key(
-                as_text(queue_key), connection=connection, job_class=job_class, serializer=serializer
+                as_text(queue_key),
+                connection=connection,
+                job_class=job_class,
+                serializer=serializer,
+                death_penalty_class=death_penalty_class,
             )
 
         all_registerd_queues = connection.smembers(cls.redis_queues_keys)
@@ -93,11 +105,12 @@ class Queue:
 
     @classmethod
     def from_queue_key(
-            cls,
-            queue_key: str,
-            connection: Optional['Redis'] = None,
-            job_class: Optional['Job'] = None,
-            serializer: Any = None,
+        cls,
+        queue_key: str,
+        connection: Optional['Redis'] = None,
+        job_class: Optional[Type['Job']] = None,
+        serializer: Any = None,
+        death_penalty_class: Optional[Type[BaseDeathPenalty]] = None,
     ) -> 'Queue':
         """Returns a Queue instance, based on the naming conventions for naming
         the internal Redis keys.  Can be used to reverse-lookup Queues by their
@@ -108,6 +121,7 @@ class Queue:
             connection (Optional[Redis], optional): Redis connection. Defaults to None.
             job_class (Optional[Job], optional): Job class. Defaults to None.
             serializer (Any, optional): Serializer. Defaults to None.
+            death_penalty_class (Optional[BaseDeathPenalty], optional): Death penalty class. Defaults to None.
 
         Raises:
             ValueError: If the queue_key doesn't start with the defined prefix
@@ -118,18 +132,37 @@ class Queue:
         prefix = cls.redis_queue_namespace_prefix
         if not queue_key.startswith(prefix):
             raise ValueError('Not a valid RQ queue key: {0}'.format(queue_key))
-        name = queue_key[len(prefix):]
-        return cls(name, connection=connection, job_class=job_class, serializer=serializer)
+        name = queue_key[len(prefix) :]
+        return cls(
+            name,
+            connection=connection,
+            job_class=job_class,
+            serializer=serializer,
+            death_penalty_class=death_penalty_class,
+        )
+
+    @classmethod
+    def get_intermediate_queue_key(cls, key: str) -> str:
+        """Returns the intermediate queue key for a given queue key.
+
+        Args:
+            key (str): The queue key
+
+        Returns:
+            str: The intermediate queue key
+        """
+        return f'{key}:intermediate'
 
     def __init__(
-            self,
-            name: str = 'default',
-            default_timeout: Optional[int] = None,
-            connection: Optional['Redis'] = None,
-            is_async: bool = True,
-            job_class: Union[str, Type['Job'], None] = None,
-            serializer: Any = None,
-            **kwargs,
+        self,
+        name: str = 'default',
+        default_timeout: Optional[int] = None,
+        connection: Optional['Redis'] = None,
+        is_async: bool = True,
+        job_class: Optional[Union[str, Type['Job']]] = None,
+        serializer: Any = None,
+        death_penalty_class: Type[BaseDeathPenalty] = UnixSignalDeathPenalty,
+        **kwargs,
     ):
         """Initializes a Queue object.
 
@@ -139,10 +172,13 @@ class Queue:
             connection (Optional[Redis], optional): Redis connection. Defaults to None.
             is_async (bool, optional): Whether jobs should run "async" (using the worker).
                 If `is_async` is false, jobs will run on the same process from where it was called. Defaults to True.
-            job_class (Union[str, 'Job', optional): Job class or a string referencing the Job class path. Defaults to None.
+            job_class (Union[str, 'Job', optional): Job class or a string referencing the Job class path.
+                Defaults to None.
             serializer (Any, optional): Serializer. Defaults to None.
+            death_penalty_class (Type[BaseDeathPenalty, optional): Job class or a string referencing the Job class path.
+                Defaults to UnixSignalDeathPenalty.
         """
-        self.connection = resolve_connection(connection)
+        self.connection = connection or resolve_connection()
         prefix = self.redis_queue_namespace_prefix
         self.name = name
         self._key = '{0}{1}'.format(prefix, name)
@@ -159,6 +195,7 @@ class Queue:
             if isinstance(job_class, str):
                 job_class = import_attribute(job_class)
             self.job_class = job_class
+        self.death_penalty_class = death_penalty_class
 
         self.serializer = resolve_serializer(serializer)
         self.redis_server_version: Optional[Tuple[int, int, int]] = None
@@ -166,9 +203,6 @@ class Queue:
     def __len__(self):
         return self.count
 
-    def __nonzero__(self):
-        return True
-
     def __bool__(self):
         return True
 
@@ -190,6 +224,11 @@ class Queue:
         """Returns the Redis key for this Queue."""
         return self._key
 
+    @property
+    def intermediate_queue_key(self):
+        """Returns the Redis key for intermediate queue."""
+        return self.get_intermediate_queue_key(self._key)
+
     @property
     def registry_cleaning_key(self):
         """Redis key used to indicate this queue has been cleaned."""
@@ -198,10 +237,11 @@ class Queue:
     @property
     def scheduler_pid(self) -> int:
         from rq.scheduler import RQScheduler
+
         pid = self.connection.get(RQScheduler.get_locking_key(self.name))
         return int(pid.decode()) if pid is not None else None
 
-    def acquire_cleaning_lock(self) -> bool:
+    def acquire_maintenance_lock(self) -> bool:
         """Returns a boolean indicating whether a lock to clean this queue
         is acquired. A lock expires in 899 seconds (15 minutes - 1 second)
 
@@ -270,7 +310,7 @@ class Queue:
         return self.count == 0
 
     @property
-    def is_async(self):
+    def is_async(self) -> bool:
         """Returns whether the current queue is async."""
         return bool(self._is_async)
 
@@ -336,7 +376,7 @@ class Queue:
         else:
             end = length
         job_ids = [as_text(job_id) for job_id in self.connection.lrange(self.key, start, end)]
-        self.log.debug(f"Getting jobs for queue {green(self.name)}: {len(job_ids)} found.")
+        self.log.debug('Getting jobs for queue %s: %d found.', green(self.name), len(job_ids))
         return job_ids
 
     def get_jobs(self, offset: int = 0, length: int = -1) -> List['Job']:
@@ -420,7 +460,7 @@ class Queue:
         Returns:
             _type_: _description_
         """
-        job_id = job_or_id.id if isinstance(job_or_id, self.job_class) else job_or_id
+        job_id: str = job_or_id.id if isinstance(job_or_id, self.job_class) else job_or_id
 
         if pipeline is not None:
             return pipeline.lrem(self.key, 1, job_id)
@@ -435,10 +475,10 @@ class Queue:
 
         self.connection.rename(self.key, COMPACT_QUEUE)
         while True:
-            job_id = as_text(self.connection.lpop(COMPACT_QUEUE))
+            job_id = self.connection.lpop(COMPACT_QUEUE)
             if job_id is None:
                 break
-            if self.job_class.exists(job_id, self.connection):
+            if self.job_class.exists(as_text(job_id), self.connection):
                 self.connection.rpush(self.key, job_id)
 
     def push_job_id(self, job_id: str, pipeline: Optional['Pipeline'] = None, at_front: bool = False):
@@ -451,30 +491,33 @@ class Queue:
             at_front (bool, optional): Whether to push the job to front of the queue. Defaults to False.
         """
         connection = pipeline if pipeline is not None else self.connection
-        if at_front:
-            result = connection.lpush(self.key, job_id)
+        push = connection.lpush if at_front else connection.rpush
+        result = push(self.key, job_id)
+        if pipeline is None:
+            self.log.debug('Pushed job %s into %s, %s job(s) are in queue.', blue(job_id), green(self.name), result)
         else:
-            result = connection.rpush(self.key, job_id)
-        self.log.debug(f"Pushed job {blue(job_id)} into {green(self.name)}, {result} job(s) are in queue.")
+            # Pipelines do not return the number of jobs in the queue.
+            self.log.debug('Pushed job %s into %s', blue(job_id), green(self.name))
 
     def create_job(
-            self,
-            func: 'FunctionReferenceType',
-            args: Union[Tuple, List, None] = None,
-            kwargs: Optional[Dict] = None,
-            timeout: Optional[int] = None,
-            result_ttl: Optional[int] = None,
-            ttl: Optional[int] = None,
-            failure_ttl: Optional[int] = None,
-            description: Optional[str] = None,
-            depends_on: Optional['JobDependencyType'] = None,
-            job_id: Optional[str] = None,
-            meta: Optional[Dict] = None,
-            status: JobStatus = JobStatus.QUEUED,
-            retry: Optional['Retry'] = None,
-            *,
-            on_success: Optional[Callable] = None,
-            on_failure: Optional[Callable] = None,
+        self,
+        func: 'FunctionReferenceType',
+        args: Union[Tuple, List, None] = None,
+        kwargs: Optional[Dict] = None,
+        timeout: Optional[int] = None,
+        result_ttl: Optional[int] = None,
+        ttl: Optional[int] = None,
+        failure_ttl: Optional[int] = None,
+        description: Optional[str] = None,
+        depends_on: Optional['JobDependencyType'] = None,
+        job_id: Optional[str] = None,
+        meta: Optional[Dict] = None,
+        status: JobStatus = JobStatus.QUEUED,
+        retry: Optional['Retry'] = None,
+        *,
+        on_success: Optional[Union[Callback, Callable]] = None,
+        on_failure: Optional[Union[Callback, Callable]] = None,
+        on_stopped: Optional[Union[Callback, Callable]] = None,
     ) -> Job:
         """Creates a job based on parameters given
 
@@ -492,8 +535,13 @@ class Queue:
             meta (Optional[Dict], optional): Job metadata. Defaults to None.
             status (JobStatus, optional): Job status. Defaults to JobStatus.QUEUED.
             retry (Optional[Retry], optional): The Retry Object. Defaults to None.
-            on_success (Optional[Callable], optional): On success callable. Defaults to None.
-            on_failure (Optional[Callable], optional): On failure callable. Defaults to None.
+            on_success (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on success. Defaults to
+                None. Callable is deprecated.
+            on_failure (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on failure. Defaults to
+                None. Callable is deprecated.
+            on_stopped (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on stopped. Defaults to
+                None. Callable is deprecated.
+            pipeline (Optional[Pipeline], optional): The Redis Pipeline. Defaults to None.
 
         Raises:
             ValueError: If the timeout is 0
@@ -534,6 +582,7 @@ class Queue:
             serializer=self.serializer,
             on_success=on_success,
             on_failure=on_failure,
+            on_stopped=on_stopped,
         )
 
         if retry:
@@ -600,23 +649,24 @@ class Queue:
         return job
 
     def enqueue_call(
-            self,
-            func: 'FunctionReferenceType',
-            args: Union[Tuple, List, None] = None,
-            kwargs: Optional[Dict] = None,
-            timeout: Optional[int] = None,
-            result_ttl: Optional[int] = None,
-            ttl: Optional[int] = None,
-            failure_ttl: Optional[int] = None,
-            description: Optional[str] = None,
-            depends_on: Optional['JobDependencyType'] = None,
-            job_id: Optional[str] = None,
-            at_front: bool = False,
-            meta: Optional[Dict] = None,
-            retry: Optional['Retry'] = None,
-            on_success: Optional[Callable[..., Any]] = None,
-            on_failure: Optional[Callable[..., Any]] = None,
-            pipeline: Optional['Pipeline'] = None,
+        self,
+        func: 'FunctionReferenceType',
+        args: Union[Tuple, List, None] = None,
+        kwargs: Optional[Dict] = None,
+        timeout: Optional[int] = None,
+        result_ttl: Optional[int] = None,
+        ttl: Optional[int] = None,
+        failure_ttl: Optional[int] = None,
+        description: Optional[str] = None,
+        depends_on: Optional['JobDependencyType'] = None,
+        job_id: Optional[str] = None,
+        at_front: bool = False,
+        meta: Optional[Dict] = None,
+        retry: Optional['Retry'] = None,
+        on_success: Optional[Union[Callback, Callable[..., Any]]] = None,
+        on_failure: Optional[Union[Callback, Callable[..., Any]]] = None,
+        on_stopped: Optional[Union[Callback, Callable[..., Any]]] = None,
+        pipeline: Optional['Pipeline'] = None,
     ) -> Job:
         """Creates a job to represent the delayed function call and enqueues it.
 
@@ -638,8 +688,12 @@ class Queue:
             at_front (bool, optional): Whether to enqueue the job at the front. Defaults to False.
             meta (Optional[Dict], optional): Metadata to attach to the job. Defaults to None.
             retry (Optional[Retry], optional): Retry object. Defaults to None.
-            on_success (Optional[Callable[..., Any]], optional): Callable for on success. Defaults to None.
-            on_failure (Optional[Callable[..., Any]], optional): Callable for on failure. Defaults to None.
+            on_success (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on success. Defaults to
+                None. Callable is deprecated.
+            on_failure (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on failure. Defaults to
+                None. Callable is deprecated.
+            on_stopped (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on stopped. Defaults to
+                None. Callable is deprecated.
             pipeline (Optional[Pipeline], optional): The Redis Pipeline. Defaults to None.
 
         Returns:
@@ -662,30 +716,28 @@ class Queue:
             retry=retry,
             on_success=on_success,
             on_failure=on_failure,
+            on_stopped=on_stopped,
         )
-
-        job = self.setup_dependencies(job, pipeline=pipeline)
-        # If we do not depend on an unfinished job, enqueue the job.
-        if job.get_status(refresh=False) != JobStatus.DEFERRED:
-            return self.enqueue_job(job, pipeline=pipeline, at_front=at_front)
-        return job
+        return self.enqueue_job(job, pipeline=pipeline, at_front=at_front)
 
     @staticmethod
     def prepare_data(
-            func: 'FunctionReferenceType',
-            args: Union[Tuple, List, None] = None,
-            kwargs: Optional[Dict] = None,
-            timeout: Optional[int] = None,
-            result_ttl: Optional[int] = None,
-            ttl: Optional[int] = None,
-            failure_ttl: Optional[int] = None,
-            description: Optional[str] = None,
-            job_id: Optional[str] = None,
-            at_front: bool = False,
-            meta: Optional[Dict] = None,
-            retry: Optional['Retry'] = None,
-            on_success: Optional[Callable] = None,
-            on_failure: Optional[Callable] = None,
+        func: 'FunctionReferenceType',
+        args: Union[Tuple, List, None] = None,
+        kwargs: Optional[Dict] = None,
+        timeout: Optional[int] = None,
+        result_ttl: Optional[int] = None,
+        ttl: Optional[int] = None,
+        failure_ttl: Optional[int] = None,
+        description: Optional[str] = None,
+        depends_on: Optional[List] = None,
+        job_id: Optional[str] = None,
+        at_front: bool = False,
+        meta: Optional[Dict] = None,
+        retry: Optional['Retry'] = None,
+        on_success: Optional[Union[Callback, Callable]] = None,
+        on_failure: Optional[Union[Callback, Callable]] = None,
+        on_stopped: Optional[Union[Callback, Callable]] = None,
     ) -> EnqueueData:
         """Need this till support dropped for python_version < 3.7, where defaults can be specified for named tuples
         And can keep this logic within EnqueueData
@@ -699,12 +751,17 @@ class Queue:
             ttl (Optional[int], optional): Time to live. Defaults to None.
             failure_ttl (Optional[int], optional): Failure time to live. Defaults to None.
             description (Optional[str], optional): The job description. Defaults to None.
+            depends_on (Optional[JobDependencyType], optional): The job dependencies. Defaults to None.
             job_id (Optional[str], optional): The job ID. Defaults to None.
             at_front (bool, optional): Whether to enqueue the job at the front. Defaults to False.
             meta (Optional[Dict], optional): Metadata to attach to the job. Defaults to None.
             retry (Optional[Retry], optional): Retry object. Defaults to None.
-            on_success (Optional[Callable[..., Any]], optional): Callable for on success. Defaults to None.
-            on_failure (Optional[Callable[..., Any]], optional): Callable for on failure. Defaults to None.
+            on_success (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on success. Defaults to
+                None. Callable is deprecated.
+            on_failure (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on failure. Defaults to
+                None. Callable is deprecated.
+            on_stopped (Optional[Union[Callback, Callable[..., Any]]], optional): Callback for on stopped. Defaults to
+                None. Callable is deprecated.
 
         Returns:
             EnqueueData: The EnqueueData
@@ -718,12 +775,14 @@ class Queue:
             ttl,
             failure_ttl,
             description,
+            depends_on,
             job_id,
             at_front,
             meta,
             retry,
             on_success,
             on_failure,
+            on_stopped,
         )
 
     def enqueue_many(self, job_datas: List['EnqueueData'], pipeline: Optional['Pipeline'] = None) -> List[Job]:
@@ -738,33 +797,66 @@ class Queue:
             List[Job]: A list of enqueued jobs
         """
         pipe = pipeline if pipeline is not None else self.connection.pipeline()
-        jobs = [
-            self.enqueue_job(
-                self.create_job(
-                    job_data.func,
-                    args=job_data.args,
-                    kwargs=job_data.kwargs,
-                    result_ttl=job_data.result_ttl,
-                    ttl=job_data.ttl,
-                    failure_ttl=job_data.failure_ttl,
-                    description=job_data.description,
-                    depends_on=None,
-                    job_id=job_data.job_id,
-                    meta=job_data.meta,
-                    status=JobStatus.QUEUED,
-                    timeout=job_data.timeout,
-                    retry=job_data.retry,
-                    on_success=job_data.on_success,
-                    on_failure=job_data.on_failure,
-                ),
-                pipeline=pipe,
-                at_front=job_data.at_front,
+        jobs_without_dependencies = []
+        jobs_with_unmet_dependencies = []
+        jobs_with_met_dependencies = []
+
+        def get_job_kwargs(job_data, initial_status):
+            return {
+                "func": job_data.func,
+                "args": job_data.args,
+                "kwargs": job_data.kwargs,
+                "result_ttl": job_data.result_ttl,
+                "ttl": job_data.ttl,
+                "failure_ttl": job_data.failure_ttl,
+                "description": job_data.description,
+                "depends_on": job_data.depends_on,
+                "job_id": job_data.job_id,
+                "meta": job_data.meta,
+                "status": initial_status,
+                "timeout": job_data.timeout,
+                "retry": job_data.retry,
+                "on_success": job_data.on_success,
+                "on_failure": job_data.on_failure,
+            }
+
+        # Enqueue jobs without dependencies
+        job_datas_without_dependencies = [job_data for job_data in job_datas if not job_data.depends_on]
+        if job_datas_without_dependencies:
+            jobs_without_dependencies = [
+                self._enqueue_job(
+                    self.create_job(**get_job_kwargs(job_data, JobStatus.QUEUED)),
+                    pipeline=pipe,
+                    at_front=job_data.at_front,
+                )
+                for job_data in job_datas_without_dependencies
+            ]
+            if pipeline is None:
+                pipe.execute()
+
+        job_datas_with_dependencies = [job_data for job_data in job_datas if job_data.depends_on]
+        if job_datas_with_dependencies:
+            # Save all jobs with dependencies as deferred
+            jobs_with_dependencies = [
+                self.create_job(**get_job_kwargs(job_data, JobStatus.DEFERRED))
+                for job_data in job_datas_with_dependencies
+            ]
+            for job in jobs_with_dependencies:
+                job.save(pipeline=pipe)
+            if pipeline is None:
+                pipe.execute()
+
+            # Enqueue the jobs whose dependencies have been met
+            jobs_with_met_dependencies, jobs_with_unmet_dependencies = Dependency.get_jobs_with_met_dependencies(
+                jobs_with_dependencies, pipeline=pipe
             )
-            for job_data in job_datas
-        ]
-        if pipeline is None:
-            pipe.execute()
-        return jobs
+            jobs_with_met_dependencies = [
+                self._enqueue_job(job, pipeline=pipe, at_front=job.enqueue_at_front)
+                for job in jobs_with_met_dependencies
+            ]
+            if pipeline is None:
+                pipe.execute()
+        return jobs_without_dependencies + jobs_with_unmet_dependencies + jobs_with_met_dependencies
 
     def run_job(self, job: 'Job') -> Job:
         """Run the job
@@ -817,6 +909,7 @@ class Queue:
         retry = kwargs.pop('retry', None)
         on_success = kwargs.pop('on_success', None)
         on_failure = kwargs.pop('on_failure', None)
+        on_stopped = kwargs.pop('on_stopped', None)
         pipeline = kwargs.pop('pipeline', None)
 
         if 'args' in kwargs or 'kwargs' in kwargs:
@@ -838,6 +931,7 @@ class Queue:
             retry,
             on_success,
             on_failure,
+            on_stopped,
             pipeline,
             args,
             kwargs,
@@ -869,6 +963,7 @@ class Queue:
             retry,
             on_success,
             on_failure,
+            on_stopped,
             pipeline,
             args,
             kwargs,
@@ -890,6 +985,7 @@ class Queue:
             retry=retry,
             on_success=on_success,
             on_failure=on_failure,
+            on_stopped=on_stopped,
             pipeline=pipeline,
         )
 
@@ -917,6 +1013,7 @@ class Queue:
             retry,
             on_success,
             on_failure,
+            on_stopped,
             pipeline,
             args,
             kwargs,
@@ -937,6 +1034,7 @@ class Queue:
             retry=retry,
             on_success=on_success,
             on_failure=on_failure,
+            on_stopped=on_stopped,
         )
         if at_front:
             job.enqueue_at_front = True
@@ -980,7 +1078,25 @@ class Queue:
         return self.enqueue_at(datetime.now(timezone.utc) + time_delta, func, *args, **kwargs)
 
     def enqueue_job(self, job: 'Job', pipeline: Optional['Pipeline'] = None, at_front: bool = False) -> Job:
-        """Enqueues a job for delayed execution.
+        """Enqueues a job for delayed execution checking dependencies.
+
+        Args:
+            job (Job): The job to enqueue
+            pipeline (Optional[Pipeline], optional): The Redis pipeline to use. Defaults to None.
+            at_front (bool, optional): Whether should enqueue at the front of the queue. Defaults to False.
+
+        Returns:
+            Job: The enqued job
+        """
+        job.origin = self.name
+        job = self.setup_dependencies(job, pipeline=pipeline)
+        # If we do not depend on an unfinished job, enqueue the job.
+        if job.get_status(refresh=False) != JobStatus.DEFERRED:
+            return self._enqueue_job(job, pipeline=pipeline, at_front=at_front)
+        return job
+
+    def _enqueue_job(self, job: 'Job', pipeline: Optional['Pipeline'] = None, at_front: bool = False) -> Job:
+        """Enqueues a job for delayed execution without checking dependencies.
 
         If Queue is instantiated with is_async=False, job is executed immediately.
 
@@ -1048,7 +1164,7 @@ class Queue:
         return job
 
     def enqueue_dependents(
-            self, job: 'Job', pipeline: Optional['Pipeline'] = None, exclude_job_id: Optional[str] = None
+        self, job: 'Job', pipeline: Optional['Pipeline'] = None, exclude_job_id: Optional[str] = None
     ):
         """Enqueues all jobs in the given job's dependents set and clears it.
 
@@ -1067,7 +1183,6 @@ class Queue:
         dependents_key = job.dependents_key
 
         while True:
-
             try:
                 # if a pipeline is passed, the caller is responsible for calling WATCH
                 # to ensure all jobs are enqueued
@@ -1086,7 +1201,7 @@ class Queue:
                         dependent_job_ids, connection=self.connection, serializer=self.serializer
                     )
                     if dependent_job
-                       and dependent_job.dependencies_are_met(
+                    and dependent_job.dependencies_are_met(
                         parent_job=job,
                         pipeline=pipe,
                         exclude_job_id=exclude_job_id,
@@ -1107,10 +1222,10 @@ class Queue:
                     registry.remove(dependent, pipeline=pipe)
 
                     if dependent.origin == self.name:
-                        self.enqueue_job(dependent, pipeline=pipe, at_front=enqueue_at_front)
+                        self._enqueue_job(dependent, pipeline=pipe, at_front=enqueue_at_front)
                     else:
                         queue = self.__class__(name=dependent.origin, connection=self.connection)
-                        queue.enqueue_job(dependent, pipeline=pipe, at_front=enqueue_at_front)
+                        queue._enqueue_job(dependent, pipeline=pipe, at_front=enqueue_at_front)
 
                 # Only delete dependents_key if all dependents have been enqueued
                 if len(jobs_to_enqueue) == len(dependent_job_ids):
@@ -1140,9 +1255,9 @@ class Queue:
         return as_text(self.connection.lpop(self.key))
 
     @classmethod
-    def lpop(cls, queue_keys: List[str], timeout: int, connection: Optional['Redis'] = None):
-        """Helper method.  Intermediate method to abstract away from some
-        Redis API details, where LPOP accepts only a single key, whereas BLPOP
+    def lpop(cls, queue_keys: List[str], timeout: Optional[int], connection: Optional['Redis'] = None):
+        """Helper method to abstract away from some Redis API details
+        where LPOP accepts only a single key, whereas BLPOP
         accepts multiple.  So if we want the non-blocking LPOP, we need to
         iterate over all queues, do individual LPOPs, and return the result.
 
@@ -1155,7 +1270,7 @@ class Queue:
 
         Args:
             queue_keys (_type_): _description_
-            timeout (int): _description_
+            timeout (Optional[int]): _description_
             connection (Optional[Redis], optional): _description_. Defaults to None.
 
         Raises:
@@ -1165,15 +1280,15 @@ class Queue:
         Returns:
             _type_: _description_
         """
-        connection = resolve_connection(connection)
+        connection = connection or resolve_connection()
         if timeout is not None:  # blocking variant
             if timeout == 0:
                 raise ValueError('RQ does not support indefinite timeouts. Please pick a timeout value > 0')
-            colored_queues = ''.join(map(str, [green(str(queue)) for queue in queue_keys]))
+            colored_queues = ', '.join(map(str, [green(str(queue)) for queue in queue_keys]))
             logger.debug(f"Starting BLPOP operation for queues {colored_queues} with timeout of {timeout}")
             result = connection.blpop(queue_keys, timeout)
             if result is None:
-                logger.debug(f"BLPOP Timeout, no jobs found on queues {colored_queues}")
+                logger.debug(f"BLPOP timeout, no jobs found on queues {colored_queues}")
                 raise DequeueTimeout(timeout, queue_keys)
             queue_key, job_id = result
             return queue_key, job_id
@@ -1184,14 +1299,36 @@ class Queue:
                     return queue_key, blob
             return None
 
+    @classmethod
+    def lmove(cls, connection: 'Redis', queue_key: str, timeout: Optional[int]):
+        """Similar to lpop, but accepts only a single queue key and immediately pushes
+        the result to an intermediate queue.
+        """
+        if timeout is not None:  # blocking variant
+            if timeout == 0:
+                raise ValueError('RQ does not support indefinite timeouts. Please pick a timeout value > 0')
+            colored_queue = green(queue_key)
+            logger.debug(f"Starting BLMOVE operation for {colored_queue} with timeout of {timeout}")
+            result = connection.blmove(queue_key, cls.get_intermediate_queue_key(queue_key), timeout)
+            if result is None:
+                logger.debug(f"BLMOVE timeout, no jobs found on {colored_queue}")
+                raise DequeueTimeout(timeout, queue_key)
+            return queue_key, result
+        else:  # non-blocking variant
+            result = connection.lmove(queue_key, cls.get_intermediate_queue_key(queue_key))
+            if result is not None:
+                return queue_key, result
+            return None
+
     @classmethod
     def dequeue_any(
-            cls,
-            queues: List['Queue'],
-            timeout: int,
-            connection: Optional['Redis'] = None,
-            job_class: Optional['Job'] = None,
-            serializer: Any = None,
+        cls,
+        queues: List['Queue'],
+        timeout: Optional[int],
+        connection: Optional['Redis'] = None,
+        job_class: Optional['Job'] = None,
+        serializer: Any = None,
+        death_penalty_class: Optional[Type[BaseDeathPenalty]] = None,
     ) -> Tuple['Job', 'Queue']:
         """Class method returning the job_class instance at the front of the given
         set of Queues, where the order of the queues is important.
@@ -1205,10 +1342,11 @@ class Queue:
 
         Args:
             queues (List[Queue]): List of queue objects
-            timeout (int): Timeout for the LPOP
+            timeout (Optional[int]): Timeout for the LPOP
             connection (Optional[Redis], optional): Redis Connection. Defaults to None.
-            job_class (Optional[Job], optional): The job classification. Defaults to None.
+            job_class (Optional[Type[Job]], optional): The job class. Defaults to None.
             serializer (Any, optional): Serializer to use. Defaults to None.
+            death_penalty_class (Optional[Type[BaseDeathPenalty]], optional): The death penalty class. Defaults to None.
 
         Raises:
             e: Any exception
@@ -1220,11 +1358,20 @@ class Queue:
 
         while True:
             queue_keys = [q.key for q in queues]
-            result = cls.lpop(queue_keys, timeout, connection=connection)
+            if len(queue_keys) == 1 and get_version(connection) >= (6, 2, 0):
+                result = cls.lmove(connection, queue_keys[0], timeout)
+            else:
+                result = cls.lpop(queue_keys, timeout, connection=connection)
             if result is None:
                 return None
             queue_key, job_id = map(as_text, result)
-            queue = cls.from_queue_key(queue_key, connection=connection, job_class=job_class, serializer=serializer)
+            queue = cls.from_queue_key(
+                queue_key,
+                connection=connection,
+                job_class=job_class,
+                serializer=serializer,
+                death_penalty_class=death_penalty_class,
+            )
             try:
                 job = job_class.fetch(job_id, connection=connection, serializer=serializer)
             except NoSuchJobError:
