@@ -7,18 +7,18 @@ terminal colorizing code, originally by Georg Brandl.
 
 import calendar
 import datetime
+import datetime as dt
 import importlib
 import logging
 import numbers
-import sys
-import datetime as dt
 from collections.abc import Iterable
-from typing import TYPE_CHECKING, Dict, List, Optional, Any, Callable, Tuple, Union
-
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union
 
 if TYPE_CHECKING:
     from redis import Redis
 
+    from .queue import Queue
+
 from redis.exceptions import ResponseError
 
 from .exceptions import TimeoutFormatError
@@ -26,106 +26,6 @@ from .exceptions import TimeoutFormatError
 logger = logging.getLogger(__name__)
 
 
-class _Colorizer:
-    def __init__(self):
-        esc = "\x1b["
-
-        self.codes = {}
-        self.codes[""] = ""
-        self.codes["reset"] = esc + "39;49;00m"
-
-        self.codes["bold"] = esc + "01m"
-        self.codes["faint"] = esc + "02m"
-        self.codes["standout"] = esc + "03m"
-        self.codes["underline"] = esc + "04m"
-        self.codes["blink"] = esc + "05m"
-        self.codes["overline"] = esc + "06m"
-
-        dark_colors = ["black", "darkred", "darkgreen", "brown", "darkblue", "purple", "teal", "lightgray"]
-        light_colors = ["darkgray", "red", "green", "yellow", "blue", "fuchsia", "turquoise", "white"]
-
-        x = 30
-        for d, l in zip(dark_colors, light_colors):
-            self.codes[d] = esc + "%im" % x
-            self.codes[l] = esc + "%i;01m" % x
-            x += 1
-
-        del d, l, x
-
-        self.codes["darkteal"] = self.codes["turquoise"]
-        self.codes["darkyellow"] = self.codes["brown"]
-        self.codes["fuscia"] = self.codes["fuchsia"]
-        self.codes["white"] = self.codes["bold"]
-
-        if hasattr(sys.stdout, "isatty"):
-            self.notty = not sys.stdout.isatty()
-        else:
-            self.notty = True
-
-    def reset_color(self):
-        return self.codes["reset"]
-
-    def colorize(self, color_key, text):
-        if self.notty:
-            return text
-        else:
-            return self.codes[color_key] + text + self.codes["reset"]
-
-
-colorizer = _Colorizer()
-
-
-def make_colorizer(color: str):
-    """Creates a function that colorizes text with the given color.
-
-    For example::
-
-        ..codeblock::python
-
-            >>> green = make_colorizer('darkgreen')
-            >>> red = make_colorizer('red')
-            >>>
-            >>> # You can then use:
-            >>> print("It's either " + green('OK') + ' or ' + red('Oops'))
-    """
-
-    def inner(text):
-        return colorizer.colorize(color, text)
-
-    return inner
-
-
-class ColorizingStreamHandler(logging.StreamHandler):
-
-    levels = {
-        logging.WARNING: make_colorizer('darkyellow'),
-        logging.ERROR: make_colorizer('darkred'),
-        logging.CRITICAL: make_colorizer('darkred'),
-    }
-
-    def __init__(self, exclude=None, *args, **kwargs):
-        self.exclude = exclude
-        super().__init__(*args, **kwargs)
-
-    @property
-    def is_tty(self):
-        isatty = getattr(self.stream, 'isatty', None)
-        return isatty and isatty()
-
-    def format(self, record):
-        message = logging.StreamHandler.format(self, record)
-        if self.is_tty:
-            colorize = self.levels.get(record.levelno, lambda x: x)
-
-            # Don't colorize any traceback
-            parts = message.split('\n', 1)
-            parts[0] = " ".join([parts[0].split(" ", 1)[0], colorize(parts[0].split(" ", 1)[1])])
-
-            message = '\n'.join(parts)
-
-        return message
-
-
 def compact(lst: List[Any]) -> List[Any]:
     """Excludes `None` values from a list-like object.
 
@@ -138,7 +38,7 @@ def compact(lst: List[Any]) -> List[Any]:
     return [item for item in lst if item is not None]
 
 
-def as_text(v: Union[bytes, str]) -> Optional[str]:
+def as_text(v: Union[bytes, str]) -> str:
     """Converts a bytes value to a string using `utf-8`.
 
     Args:
@@ -150,9 +50,7 @@ def as_text(v: Union[bytes, str]) -> Optional[str]:
     Returns:
         value (Optional[str]): Either the decoded string or None
     """
-    if v is None:
-        return None
-    elif isinstance(v, bytes):
+    if isinstance(v, bytes):
         return v.decode('utf-8')
     elif isinstance(v, str):
         return v
@@ -182,9 +80,6 @@ def import_attribute(name: str) -> Callable[..., Any]:
     E.g.: package_a.package_b.module_a.ClassA.my_static_method
 
     Thus we remove the bits from the end of the name until we can import it
-    Sometimes the failure during importing is due to a genuine coding error in the imported module
-    In this case, the exception is logged as a warning for ease of debugging.
-    The above logic will apply anyways regardless of the cause of the import error.
 
     Args:
         name (str): The name (reference) to the path.
@@ -204,24 +99,28 @@ def import_attribute(name: str) -> Callable[..., Any]:
             module = importlib.import_module(module_name)
             break
         except ImportError:
-            logger.warning("Import error for '%s'" % module_name, exc_info=True)
             attribute_bits.insert(0, module_name_bits.pop())
 
     if module is None:
-        raise ValueError('Invalid attribute name: %s' % name)
+        # maybe it's a builtin
+        try:
+            return __builtins__[name]
+        except KeyError:
+            raise ValueError('Invalid attribute name: %s' % name)
 
     attribute_name = '.'.join(attribute_bits)
     if hasattr(module, attribute_name):
         return getattr(module, attribute_name)
-
     # staticmethods
     attribute_name = attribute_bits.pop()
     attribute_owner_name = '.'.join(attribute_bits)
-    attribute_owner = getattr(module, attribute_owner_name)
+    try:
+        attribute_owner = getattr(module, attribute_owner_name)
+    except:  # noqa
+        raise ValueError('Invalid attribute name: %s' % attribute_name)
 
     if not hasattr(attribute_owner, attribute_name):
         raise ValueError('Invalid attribute name: %s' % name)
-
     return getattr(attribute_owner, attribute_name)
 
 
@@ -354,7 +253,7 @@ def str_to_date(date_str: Optional[str]) -> Union[dt.datetime, Any]:
         return utcparse(date_str.decode())
 
 
-def parse_timeout(timeout: Any):
+def parse_timeout(timeout: Union[int, float, str]) -> int:
     """Transfer all kinds of timeout format to an integer representing seconds"""
     if not isinstance(timeout, numbers.Integral) and timeout is not None:
         try:
@@ -467,3 +366,16 @@ def get_call_string(
     args = ', '.join(arg_list)
 
     return '{0}({1})'.format(func_name, args)
+
+
+def parse_names(queues_or_names: List[Union[str, 'Queue']]) -> List[str]:
+    """Given a list of strings or queues, returns queue names"""
+    from .queue import Queue
+
+    names = []
+    for queue_or_name in queues_or_names:
+        if isinstance(queue_or_name, Queue):
+            names.append(queue_or_name.name)
+        else:
+            names.append(str(queue_or_name))
+    return names
